// Copyright (C) 2016 Jochen Becher
// SPDX-License-Identifier: LicenseRef-Qt-Commercial OR GPL-3.0-only WITH Qt-GPL-exception-1.0

#include "qark/flag.h"

namespace qark {

Flag::MaskType Flag::m_nextMask = 1;

} // namespace qark
